'use strict';

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _isNativeReflectConstruct() {
  if (typeof Reflect === "undefined" || !Reflect.construct) return false;
  if (Reflect.construct.sham) return false;
  if (typeof Proxy === "function") return true;

  try {
    Date.prototype.toString.call(Reflect.construct(Date, [], function () {}));
    return true;
  } catch (e) {
    return false;
  }
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

function _createSuper(Derived) {
  return function () {
    var Super = _getPrototypeOf(Derived),
        result;

    if (_isNativeReflectConstruct()) {
      var NewTarget = _getPrototypeOf(this).constructor;

      result = Reflect.construct(Super, arguments, NewTarget);
    } else {
      result = Super.apply(this, arguments);
    }

    return _possibleConstructorReturn(this, result);
  };
}

var getRenderedMarkupLength = function getRenderedMarkupLength(markup) {
  return markup.replace(/<[^>]*>/g, '').length;
};

var Snippet = /*#__PURE__*/function () {
  function Snippet(start, content) {
    _classCallCheck(this, Snippet);

    this.start = start;
    this.content = content;
  }

  _createClass(Snippet, [{
    key: "getLength",
    value: function getLength() {
      return getRenderedMarkupLength(this.content);
    }
  }, {
    key: "getContent",
    value: function getContent() {
      return this.content;
    }
  }]);

  return Snippet;
}();

var HtmlSnippet = /*#__PURE__*/function (_Snippet) {
  _inherits(HtmlSnippet, _Snippet);

  var _super = _createSuper(HtmlSnippet);

  function HtmlSnippet(start) {
    _classCallCheck(this, HtmlSnippet);

    return _super.call(this, start);
  }

  _createClass(HtmlSnippet, [{
    key: "getOpening",
    value: function getOpening() {
      return this.content.match(new RegExp("^<".concat(this.name, "[^>]*>")))[0];
    }
  }, {
    key: "getClosing",
    value: function getClosing() {
      return "</".concat(this.name, ">");
    }
  }, {
    key: "getInner",
    value: function getInner() {
      return this.content.replace(new RegExp("^<".concat(this.name, "[^>]*>")), '').replace(new RegExp("</".concat(this.name, ">$")), '');
    }
  }]);

  return HtmlSnippet;
}(Snippet);

var TextSnippet = /*#__PURE__*/function (_Snippet) {
  _inherits(TextSnippet, _Snippet);

  var _super = _createSuper(TextSnippet);

  function TextSnippet(start, content) {
    var _this;

    _classCallCheck(this, TextSnippet);

    _this = _super.call(this, start, content); // TODO may have sense this: this.end = start + content.length;

    _this.end = start + 1;
    return _this;
  }

  return TextSnippet;
}(Snippet);

var getRowsFromText = function getRowsFromText(snippet, maxLength, spaceLeft) {
  var words = snippet.content.split(' ');
  return words.reduce(function (rows, word, index) {
    if (index < words.length - 1) {
      word = "".concat(word, " "); // <- we must add the space...
    }

    if (rows.length === 0) {
      if (word.length <= spaceLeft) {
        rows.push(word);
      }
    } else {
      var lastIndex = rows.length - 1;
      var lastRow = rows[lastIndex];

      if (lastRow.length + word.length <= spaceLeft) {
        rows[lastIndex] = "".concat(lastRow).concat(word);
      } else {
        rows.push(word);
        spaceLeft = maxLength;
      }
    }

    return rows;
  }, []);
};

var addTextSnippet = function addTextSnippet(markup, i, snippets) {
  snippets.push(new TextSnippet(i, markup[i]));
  var lastSnippet = snippets.slice(-1)[0];
  i++;

  while (markup[i] !== '<' && i < markup.length) {
    lastSnippet.content += markup[i];
    lastSnippet.end = lastSnippet.start + lastSnippet.content.length;
    ++i;
  }

  return i;
};

var addHtmlSnippet = function addHtmlSnippet(markup, i, snippets) {
  snippets.push(new HtmlSnippet(i));
  var lastSnippet = snippets[snippets.length - 1];
  ++i;
  var name = '';

  while (markup[i] !== '>' && markup[i] !== ' ' && i < markup.length) {
    name += markup[i];
    lastSnippet.name = name;
    ++i;
  }

  var leftContent, rightContent; // Handle tag with attributes

  if (markup[i] === ' ') {
    leftContent = markup.substr(lastSnippet.start, Infinity).match(new RegExp("<".concat(name, "\\s+[^>]+>")))[0];
    rightContent = markup.substr(lastSnippet.start + leftContent.length, Infinity).match(new RegExp("[^<]*</".concat(name, ">")))[0];
  } else if (markup[i] === '>') {
    leftContent = markup.substr(lastSnippet.start, i - lastSnippet.start + 1);
    var stopPoint = Math.random().toString().split('.')[1];

    if (markup.substr(i + 1, Infinity).match(new RegExp("</".concat(name, ">")))) {
      rightContent = markup.substr(i + 1, Infinity).replace("</".concat(name, ">"), "</".concat(stopPoint, ">")).match(new RegExp(".*/".concat(stopPoint, ">")))[0].replace("</".concat(stopPoint, ">"), "</".concat(name, ">"));
    } else {
      rightContent = markup.substr(i + 1, Infinity) + "</".concat(name, ">");
    }
  }

  lastSnippet.content = leftContent + rightContent;
  lastSnippet.end = lastSnippet.start + lastSnippet.content.length;
  return lastSnippet.end;
};

var getMarkupSnippets = function getMarkupSnippets(markup) {
  var snippets = [];
  var i = 0;

  while (i < markup.length) {
    if (markup[i] === '<') {
      i = addHtmlSnippet(markup, i, snippets);
    } else {
      i = addTextSnippet(markup, i, snippets);
    }
  }

  return snippets;
};

var blockElements = ['div', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'li', 'ul'];

var isBlockElement = function isBlockElement(tagName) {
  return blockElements.includes(tagName);
};

var getRowsFromHtml = function getRowsFromHtml(snippet, maxLength, spaceLeft) {
  if (!isBlockElement(snippet.name)) {
    var left = snippet.getOpening();
    var right = snippet.getClosing();
    var innerSnippets = getMarkupSnippets(snippet.getInner());
    var innerSnippet = innerSnippets[0];

    while (innerSnippets.length === 1 && innerSnippet instanceof HtmlSnippet) {
      left += innerSnippet.getOpening();
      right = innerSnippet.getClosing() + right;
      innerSnippets = getMarkupSnippets(innerSnippet.getInner());
      innerSnippet = innerSnippets[0];
    }

    var rows = splitRowByMaxLength(innerSnippets, maxLength);
    rows[0] = left + rows[0];
    rows[rows.length - 1] += right;
    return rows;
  } else {
    return [snippet.getContent()];
  }
};

var getClosingTagName = function getClosingTagName(line) {
  var lineLength = line.length;
  var i = lineLength - 2;
  var found = false;
  var tagName = '';

  while (i > 0 && !found) {
    var newChar = line[i];

    if (newChar !== '/') {
      tagName = newChar + tagName;
      --i;
    } else {
      found = true;
    }
  }

  return tagName;
};

var isLastRowBlockElement = function isLastRowBlockElement(lastRow) {
  if (lastRow.endsWith('>')) {
    var tagName = getClosingTagName(lastRow);
    return isBlockElement(tagName);
  }

  return false;
};

var canContentBeAppended = function canContentBeAppended(lastRow, lastRowLength, maxLength, snippet) {
  // New content can be appended if and only if nor the lastRow neither the new snippet are block-level elements...
  if (!isLastRowBlockElement(lastRow) && !isBlockElement(snippet.name)) {
    // ...and there is sufficient left space to add snippet's content
    return lastRowLength + snippet.getLength() <= maxLength;
  }

  return false;
};

var splitRowByMaxLength = function splitRowByMaxLength(snippets, maxLength) {
  return snippets.reduce(function (rows, snippet) {
    if (rows.length === 0) {
      if (snippet instanceof TextSnippet) {
        rows = rows.concat(getRowsFromText(snippet, maxLength, maxLength));
      } else if (snippet instanceof HtmlSnippet) {
        var rowsFromHtml = getRowsFromHtml(snippet, maxLength);
        rows = rows.concat(rowsFromHtml);
      }
    } else {
      var lastIndex = rows.length - 1;
      var lastRow = rows[lastIndex];
      var lastRowLength = getRenderedMarkupLength(lastRow);
      var lastRowIsBlockElement = isLastRowBlockElement(lastRow);

      if (canContentBeAppended(lastRow, lastRowLength, maxLength, snippet)) {
        rows[lastIndex] = "".concat(lastRow).concat(snippet.getContent());
      } else {
        if (snippet instanceof TextSnippet) {
          var rowsFromText;

          if (!lastRowIsBlockElement) {
            rowsFromText = getRowsFromText(snippet, maxLength, maxLength - lastRowLength);
            rows[lastIndex] = "".concat(lastRow).concat(rowsFromText.splice(0, 1)[0]);
          } else {
            rowsFromText = getRowsFromText(snippet, maxLength, maxLength);
          }

          rows = rows.concat(rowsFromText);
        } else if (snippet instanceof HtmlSnippet) {
          var _rowsFromHtml = getRowsFromHtml(snippet, maxLength);

          rows = rows.concat(_rowsFromHtml);
        }
      }
    }

    return rows;
  }, []);
};

var getTotalNeededSpace = function getTotalNeededSpace(allRows, maxLines) {
  return allRows.slice(0, maxLines).reduce(function (space, row) {
    return space + getRenderedMarkupLength(row);
  }, 0);
};

var getSpliceDeleteCount = function getSpliceDeleteCount(allRows, maxLines, maxLineLength) {
  var i = maxLines;
  var availableSpace = maxLines * maxLineLength;

  if (availableSpace < getTotalNeededSpace(allRows, maxLines)) {
    i = 0;

    while (i < maxLines && availableSpace >= maxLineLength) {
      var nextCandidateRow = allRows[i];
      var nextCandidateRowLength = getRenderedMarkupLength(nextCandidateRow);

      if (nextCandidateRowLength <= maxLineLength) {
        availableSpace -= maxLineLength;

        if (availableSpace > 0) {
          i++;
        }
      } else {
        // nextCandidateRowLength > maxLineLength
        while (nextCandidateRowLength > maxLineLength && availableSpace > 0) {
          nextCandidateRowLength -= maxLineLength;
          availableSpace -= maxLineLength;
        }

        if (availableSpace > 0) {
          i++;
        }
      }
    }
  }

  return i;
};

var splitMarkup = function splitMarkup(markup, maxLines, maxLineLength) {
  var snippets = getMarkupSnippets(markup.replace(/\n/g, ''));
  var allRows = splitRowByMaxLength(snippets, maxLineLength);
  var deleteCount = getSpliceDeleteCount(allRows, maxLines, maxLineLength);
  var visibleRows = allRows.splice(0, deleteCount);
  var visibleMarkup = visibleRows.join('');
  var hiddenMarkup = allRows.length ? allRows.join('') : '';
  return [visibleMarkup, hiddenMarkup];
};

module.exports = splitMarkup;
